"""
:synopsis: Optimisers to use with NPU API for training
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""


def SGD(lr=0.001, momentum=0):
    """SGD Optimiser.

          :param lr: Learning rate
          :type lr: float
          :param momentum: Momentum
          :type momentum: float
      """
    return {
        "optimiser": "SGD",
        "opt_args": {
            "lr": lr,
            "momentum": momentum
        }}


def RMS(lr=0.001, decay=0):
    """RMS Optimiser.

              :param lr: Learning rate
              :type lr: float
              :param decay: Decay
              :type decay: float
          """
    return {
        "optimiser": "RMS",
        "opt_args": {
            "lr": lr,
            "decay": decay
        }}


def Adam(lr=0.001):#, beta1=0.9, beta2=0.999, epsilon=1e-08):
    """Adam Optimiser.

              :param lr: Learning rate
              :type lr: float
              :param beta1: Beta 1
              :type beta1: float
              :param beta2: Beta 2
              :type beta2: float
              :param epsilon: Epsilon
              :type epsilon: float
          """
    return {
        "optimiser": "Adam",
        "opt_args": {
            "lr": lr
#             "beta1": beta1,
#             "beta2": beta2,
#             "epsilon": epsilon,
        }}



"""
:synopsis: Datasets class wrapper
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""


class Dataset(object):

    def __init__(self, id, train_end, val_end, test_end):
        self.id = id
        self.__train_end = train_end
        self.__val_end = val_end
        self.__test_end = test_end

    def __str__(self):
        return self.id

    def __getitem__(self, item):
        return {"id": self, "indexes": item}

    @property
    def train(self):
        """
        Property for training subset
        :return: slice
        """
        return self[0:self.__train_end]

    @property
    def val(self):
        """
        Property for validation subset
        :return: slice
        """
        return self[self.__train_end:self.__val_end]

    @property
    def test(self):
        """
        Property for test subset
        :return: slice
        """
        return self[self.__val_end:self.__test_end]



"""
:synopsis: The main npu api
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""


from .npu import api, predict, compile, train, export, upload_data
from . import vision, optim, loss
from .core import DataLoader
from .version import __version__, _module


api.__module__ = _module
predict.__module__ = _module
compile.__module__ = _module
train.__module__ = _module
export.__module__ = _module
upload_data.__module__ = _module



# @wait_task
import os
import re

import requests

from .common import getToken
from .Task import Task
from .web.urls import EXPORT_URL
from .Model import Model


def export(model, path):
    if isinstance(model, Task):
        task = Task(model.task_id, show=False)
        task.wait()
        model = task.get_result()
    return exportApi(model, path)


def exportApi(model, path):
    params = {"token": getToken()}
    if isinstance(model, Model):
        params["model_name"] = model.name
    elif isinstance(model, str):
        params["model_name"] = model
    elif model != "":
        params["modelId"] = model
    response = requests.get(EXPORT_URL, params=params, stream=True)
    if response.status_code == 200:
        d = response.headers['content-disposition']
        filename = re.findall("filename=(.+)", d)[0]
        print(filename)
        with open(os.path.join(path, filename), "wb") as file:
            for chunk in response.iter_content(chunk_size=1024):
                # writing one chunk at a time to file
                if chunk:
                    file.write(chunk)
        print("Model exported.")
    else:
        raise Exception("{}".format(response.content))
    return True





import pkg_resources

__version__ = "0.2.31"

_module = "npu"


latest_version = pkg_resources.get_distribution(_module).version
if latest_version != __version__:
    print("Current version of npu library is {}. Latest version on pypi is {}".format(__version__, latest_version))



from . import core, optim, loss


def api(token, verbosity=1):
    """Use this function to get access to the API, providing your token. All subsequents API calls will use this token.

        :param token: Token from dashboard
        :type token: str
    """
    core.api(token, verbosity)


def compile(model, input_shape=[], library="", model_label="", asynchronous=False):
    """Use this to upload and compile your model. Compatible frameworks are:

        * Pytorch
        * Tensorflow 2
        * Mxnet

        :param model: Original model to compile.
        :type model: Object from framework or filename(str) to the model. `Tensorflow` + `mxnet` models must be tarred if using filenames.
        :param input_shape: Input shape of model
        :type input_shape: List
        :param library: Library used
        :type library: str (pytorch, tf, mxnet)
        :param model_label: Label for model
        :type model_label: str, optional
        :param asynchronous: If call should run async or not. Default=`False`
        :type asynchronous: bool, optional.

        :return: compiled model.
    """
    return core.compile(model, input_shape, library, model_label, asynchronous)


def upload_data(data, name=""):
    """Use this to upload your data.

        :param data: Raw data to upload.
        :type data: (numpy, Pytorch Tensor, Mxnet NDArray, Tensorflow tf.Data)
        :param name: Name to be given to data.
        :type name: str, optional

        :return: data as id.
    """
    return core.upload_data(data, name)


def predict(model, data, asynchronous=False, callback=None, **kwargs):
    """Perform a predict using a model. Default behaviour is synchronous.

        :param model: Model used to predict
        :type model: From :func:`npu.compile` or :func:`npu.train`. Id (str) or global :class:`npu.vision.models.Model` can be used.
        :param data: Data to be used for prediction
        :type data: numpy array
        :param asynchronous: If call should run async or not. Default=`False`. If you want to get the result back explicitly, call "get_result()" on returned value.
        :type asynchronous: bool, optional.
        :param callback: runs a callback function on results (asynchronous)
        :type callback: function
    """
    return core.predict(model, data, asynchronous, callback, **kwargs)


def train(model, train_data, val_data="", batch_size=32, epochs=1, optim=optim.SGD(), loss=loss.SparseCrossEntropyLoss, trained_model_name="", asynchronous=False, callback=None, **kwargs):
    """Perform a train using a model. Default behaviour is synchronous.

        :param model: Model used to predict
        :type model: From :func:`npu.compile` or :func:`npu.train`. Id (str) or global :class:`npu.vision.models.Model` can be used.
        :param train_data: Training data in format of (x, y)
        :type train_data: numpy array
        :param val_data: Validation data in format of (x, y)
        :type val_data: numpy array
        :param batch_size: Batch size for training. Default=`32`
        :type batch_size: int, optional
        :param epochs: Epoch cycles for training. Default=`1`
        :type epochs: int, optional
        :param optim: Optimiser to use
        :type optim: :func:`npu.optim`
        :param loss: Loss function to use
        :type loss:
        :param asynchronous: If call should run async or not. Default=`False`. If you want to get the result back explicitly, call "get_result()" on returned value.
        :type asynchronous: bool, optional.
        :param callback: runs a callback function on results (asynchronous)
        :type callback: function
    """
    return core.train(model, train_data, val_data, batch_size, epochs, optim, loss, trained_model_name, asynchronous, callback, **kwargs)


def export(model, path="."):
    """Export a model to file. This will export it in the original format it is in. Global models will be exported as
    pytorch models.

        :param model: Model to export
        :type model: From :func:`npu.compile` or :func:`npu.train`. Id (str) or global :class:`npu.vision.models.Model` can be used.
        :param path: Path to where the model is saved to. Default is ".".
        :type path: str, optional
    """
    return core.export(model, path)







import time

import requests
from progress.bar import ChargingBar
from progress.spinner import PixelSpinner

from .web.urls import TASK_STATUS_URL

FAILURE = "FAILURE"
PENDING = "PENDING"
COMPLETE = "COMPLETE"
STOPPED = "STOPPED"

TASK_DONE_LIST = (FAILURE, COMPLETE, STOPPED)

import threading

bar_suffix = '%(percent)d%% || %(eta)d seconds remaining || %(elapsed)d seconds elapsed'


class Task:

    def __init__(self, task_id, callback=None, show=True):
        self.task_id = task_id
        self.task_result = ""
        self.task_type = ""
        self.progress = 0
        self.task_state = PENDING
        self.callback = callback
        self.cache = None
        self._metrics = {}
        if self.callback:
            t = threading.Thread(target=self.callback_thread)
            t.start()
        if show:
            self.update()
            print("Started {0}. View status at https://dashboard.neuro-ai.co.uk/tasks/{1}".format(self.task_type.lower(), task_id))

    def wait(self, show_progress=False):
        if show_progress:
            bar = ChargingBar('Training', suffix=bar_suffix)
        else:
            spinner = PixelSpinner('Task in Progress ')
        while not self.finished():
            time.sleep(0.1)
            if show_progress:
                bar.goto(self.progress*100)
            else:
                spinner.next()
        if self.task_state == FAILURE:
            self.update(include_result=True)
            raise Exception("ERROR for task {}: {}".format(self.task_id, self.task_result))
        if self.task_state == STOPPED:
            print("Task has been stopped.")
        if show_progress:
            bar.goto(100)
            bar.finish()
        else:
            spinner.finish()

    def callback_thread(self):
        self.get_result()
        self.callback(self)

    def get_result(self):
        self.wait()
        self.update(include_result=True)
        return self.task_result

    def update(self, include_result=False):
        from .common import getResponse
        params = {"include_result": include_result}
        response = requests.get(TASK_STATUS_URL + self.task_id, params=params)
        response = getResponse(response)
        self.task_state = response["state"]
        self.task_type = response["taskType"]
        self.progress = response["progress"]
        if "result" in response:
            self.task_result = response["result"]
        if "metrics" in response:
            self._metrics = response["metrics"]

    def __str__(self):
        return str(self.get_result())

    def finished(self):
        self.update()
        return self.task_state in TASK_DONE_LIST

    def metrics(self):
        self.get_result()
        return self._metrics





from .compile import compile
from .common import checkData as upload_data
from .predict import predict
from .train import train
from .common import api
from .Dataset import Dataset
from .Model import Model
from .export import export
from .DataLoader import DataLoader



import hashlib
import math

import requests
from bson import ObjectId
from progress.bar import ChargingBar
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from .Model import Model
from .DataLoader import DataLoader
from .Dataset import Dataset
from .Task import Task, bar_suffix
from .saving.saving import save_data, determine_model
from .web.urls import TOKEN_URL, HASH_URL, UPLOAD_DATA_URL

VERBOSITY = 1
MIN_VERBOSITY = 1
MID_VERBOSITY = 2
FULL_VERBOSITY = 3

_token = ""


def api(token_, verbosity):
    global _token
    global VERBOSITY
    _token = token_
    VERBOSITY = verbosity
    if VERBOSITY != MIN_VERBOSITY:
        print("Verbosity level set to {}".format(VERBOSITY))
    params = {"token": _token}
    response = requests.get(TOKEN_URL, params=params)
    if response.status_code == 200:
        print("Token successfully authenticated")
        return response
    else:
        raise ValueError(response.text)
    # "API token not valid"


def getToken():
    return _token


def getVerbosity():
    return VERBOSITY


def getResponse(response):
    try:
        return response.json()
    except Exception as e:
        raise ConnectionError("Invalid response received. Error: {}".format(response.text))


# https://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python
def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def add_kwargs_to_params(params, **kwargs):
    params = {**params, **kwargs}
    return params


def read_in_chunks(file_object, chunk_size=1024):
    """Lazy function (generator) to read a file piece by piece.
    Default chunk size: 1k."""
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


def checkModel(model):
    from .Task import Task
    from .Model import Model
    if not isinstance(model, Task) and not isinstance(model, str) and not isinstance(model, Model):
        raise ValueError("Model is not a valid format. Please make sure you've compiled it first.")


def check_model_type(model, params):
    if isinstance(model, Model):
        params["model_name"] = model.name
    elif isinstance(model, str) and not ObjectId.is_valid(model):
        params["model_name"] = model
    elif model != "" and not isinstance(model, Task):
        params["modelId"] = model


def check_data_type(data, param_name, params):
    if isinstance(data, Dataset):
        params[param_name + "_name"] = data.id
    elif isinstance(data, str) and not ObjectId.is_valid(data):
        params[param_name + "_name"] = data
    elif data != "" and not isinstance(data, Task):
        params[param_name + "Id"] = data


def checkData(data, name=""):
    if not isinstance(name, str):
        raise ValueError("Name given is not valid. Please supply a string.")
    if isinstance(data, (str, dict)):
        return data
    elif isinstance(data, Dataset):
        return data
    elif isinstance(data, DataLoader):
        response = uploadDataLoader(data, name)
    else:
        response = uploadData(data, name)
    status_code = response.status_code
    if status_code not in (204, 200, 201):
        raise ConnectionAbortedError("Data upload has not worked: {}".format(response.content))
    if status_code != 204:
        response = getResponse(response)
    if isinstance(response, dict) and status_code == 200:
        message = response.get("message")
        print(message)
        response = response["id"]
    return response


def sliceData(data):
    id = data["id"]
    x = data["indexes"]
    y = None
    if isinstance(x, slice):
        y = x.stop
        x = x.start
    return id, x, y


def gen(dl):
    for data_part in dl.numpy():
        yield save_data(data_part)


class UploadBar(ChargingBar):
    suffix = bar_suffix + " || %(index_bytes)s / %(max_bytes)s"

    @property
    def index_bytes(self):
        return convert_size(self.index)

    @property
    def max_bytes(self):
        return convert_size(self.max)


def create_callback(encoder):
    encoder_len = encoder.len
    bar = UploadBar('Uploading', max=encoder_len)

    def callback(monitor):
        bar.goto(monitor.bytes_read)
        if monitor.bytes_read == encoder_len:
            bar.finish()
    return callback


def get_progress_bar_uploader(file):
    encoder = create_upload(file)
    callback = create_callback(encoder)
    monitor = MultipartEncoderMonitor(encoder, callback)
    return monitor


def create_upload(file):
    return MultipartEncoder({
        'file': ('file', file, 'application/octet-stream', {'Content-Transfer-Encoding': 'binary'}),
    })


def uploadDataLoader(dl, name=""):
    length = len(dl)
    if VERBOSITY >= MID_VERBOSITY:
        print("Hashing data locally...")
    hash, size = dl.hash()
    params = {"token": getToken(), "hash": hash, "collection": 1, "chunked": True, "is_last": False, "size": size,
              "given_name": name, "input_shape": dl.shape}
    # params = {"token": getToken(), "hash": hash, "collection": 1, "size": size, "given_name": name}
    print("Checking if data is on servers...")
    response = requests.get(HASH_URL, params=params)
    if response.status_code == 200:
        if VERBOSITY >= MID_VERBOSITY:
            print("Data already uploaded. Will not reupload.")
        return response
    print("Data not on servers. Starting to upload. Total size of data is {}".format(convert_size(size)))
    if length == 1:
        return uploadData(next(dl.numpy()), name)
    print("{} chunks to upload...".format(length))
    for i, data_part in enumerate(dl.numpy()):
        if VERBOSITY >= MID_VERBOSITY:
            print("Uploading chunk {} out of {}...".format(i + 1, length))
        if i == length - 1:
            params["is_last"] = True
        file = save_data(data_part)
        monitor = get_progress_bar_uploader(file)
        response = requests.post(UPLOAD_DATA_URL, data=monitor,
                                 headers={'Content-Type': monitor.content_type}, params=params)
        # response = requests.post(UPLOAD_DATA_URL, files=files, params=params)
    return response


def uploadData(data, name=""):
    if VERBOSITY >= FULL_VERBOSITY:
        print("Saving data locally...")
    if isinstance(data, str):
        file = open(data, "rb")
    else:
        file = save_data(data)
    if VERBOSITY >= FULL_VERBOSITY:
        print("Hashing...")
    hash = hashlib.md5()
    size = 0
    for piece in read_in_chunks(file):
        size += len(piece)
        hash.update(piece)
    hash = hash.hexdigest()
    print("Checking if data is on servers...")
    params = {"token": getToken(), "hash": hash, "collection": 1, "given_name": name}
    response = requests.get(HASH_URL, params=params)
    if response.status_code == 200:
        if VERBOSITY >= MID_VERBOSITY:
            print("Data already on servers. Returning result...")
        file.close()
        return response
    print("Data not found on servers. Total size of data is {}. Uploading now...".format(convert_size(size)))
    file.seek(0)
    # files = {'file': file}

    monitor = get_progress_bar_uploader(file)
    response = requests.post(UPLOAD_DATA_URL, data=monitor,
                             headers={'Content-Type': monitor.content_type}, params=params)
    # response = requests.post(UPLOAD_DATA_URL, files=files, params=params)
    if isinstance(data, str):
        file.close()
    return response


def hashData(data):
    pass


def validate_model(model, data):
    library = determine_model(model)
    if isinstance(data, str):
        return
    # data = convert_to_numpy(data)
    if library == "pytorch":
        from torch import ones
    elif library == "mxnet":
        from mxnet import nd
        ones = nd.ones
    elif library == "TF":
        from numpy import ones
    else:
        return
        # raise ValueError("Cannot validate library: {} .".format(library))
    placeholder_data = ones(data.shape)
    model(placeholder_data)


def determine_data(data):
    start = end = None
    name = ""
    if isinstance(data, dict):
        data, start, end = sliceData(data)
    if isinstance(data, Dataset):
        name = data.id
        data = data
    return data, name, start, end



import glob
import hashlib
import tarfile
import os
from io import BytesIO
from tempfile import TemporaryFile
import numpy as np


def convert_to_numpy(data):
    if isinstance(data, (tuple, list)):
        return [framework_to_numpy(d) for d in data]
    else:
        return framework_to_numpy(data)


def save_model(model, library: str):
    model_path = model
    if not isinstance(model, str):
        if not os.path.isdir("tmp"):
            os.mkdir("tmp")
        model_path = "tmp/tmp_model"
        if library == "pytorch":
            from torch import save
            import dill
            model_path += ".pt"
            save(model, model_path, pickle_module=dill)
        elif library == "keras":
            model_path += ".h5"
            import keras
            with keras.backend.get_session().graph.as_default():
                model.save_model(model_path)
        elif library == "mxnet":
            model_path += ".tar"
            model.export(model_path)
            with tarfile.open(model_path, "w") as tar:
                jsonname = model_path + "-symbol.json"
                paramname = model_path + "-0000.params"
                tar.add(jsonname, arcname=os.path.basename(jsonname))
                tar.add(paramname, arcname=os.path.basename(paramname))
        elif library == "TF":
            from tensorflow.keras.models import save_model
            model_path += ".tar"
            model.save(model_path + "dir")
            # save_model(model, model_path + "dir", include_optimizer=True, save_format='tf')
            with tarfile.open(model_path, "w") as tar:
                for file in glob.glob(model_path + "dir/*"):
                    tar.add(file, arcname=os.path.basename(file))
        elif library == "TF1":
            model_path += ".pb"
        else:
            raise ValueError("Model type: " + str(library) + " not defined")
    return model_path

    # if model_type is ModelType.ONNX:
    #     onnx.save(model, file_path)
    # elif model_type is ModelType.TF1:
    #     pass
    # tf.saved_model.save(model, "./dd.pb")
    # tf.compat.v1.saved_model.save(model, "tmp")
    # raise ValueError("Tensorflow 1 incompatible. Please use .pb file directly if using Tensorflow 1.")


def utf_str(obj):
    return str(obj).encode("utf-8")


def hash_model(model, library):
    hash = hashlib.md5()
    if isinstance(model, str):
        with open(model, "rb") as file:
            return hashlib.md5(file.read()).hexdigest()
    elif library == "pytorch":
        hash.update(utf_str(model))
        for p in model.parameters():
            hash.update(utf_str(p))
    elif library == "mxnet":
        from mxnet import sym
        x = sym.var('data')
        x = model(x).tojson()
        hash.update(utf_str(x))
        for param in sorted(model.collect_params().items()):
            hash.update(utf_str(param[1].data()))
    elif library == "TF":
        try:
            hash.update(utf_str(model.to_json()))
            for param in model.trainable_variables:
                hash.update(utf_str(param))
        except NotImplementedError:
            return ""
    else:
        raise ValueError("Cannot hash... Model type: {} not defined.".format(library))
    return hash.hexdigest()

def framework_to_numpy(data):
    try:
        if isinstance(data, np.ndarray):
            return data
    except:
        pass
    try:
        from torch import Tensor
        if isinstance(data, Tensor):
            return data.numpy()
    except:
        pass
    try:
        from tensorflow import is_tensor, make_ndarray
        if is_tensor(data):
            return make_ndarray(data)
    except:
        pass
    try:
        from mxnet import nd
        if isinstance(data, nd.NDArray):
            return data.asnumpy()
    except:
        pass
    raise ValueError("Could not determine framework data is from. Try pass in numpy array directly.")


def save_data(data):
    file = BytesIO()
    data = convert_to_numpy(data)
    if isinstance(data, (tuple, list)):
        dict_data = {str(i): data[i] for i in range(0, len(data))}
        np.savez(file, **dict_data)
    else:
        np.savez(file, data)
    file.seek(0)
    return file


def determine_model(model):
    try:
        from torch.nn import Module
        if isinstance(model, Module):
            return "pytorch"
    except:
        pass
    try:
        from tensorflow.keras import Model
        if isinstance(model, Model):
            return "TF"
    except:
        pass
    try:
        from mxnet.gluon.nn import Block
        if isinstance(model, Block):
            return "mxnet"
    except:
        pass
    if isinstance(model, str):
        if model.endswith(".pt"):
            return "pytorch"
    raise ValueError("Could not determine framework model is from. Please specify explicitly.")



"""
:synopsis: Supported loss functions to use with NPU API for training.
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""

# CrossEntropyLoss = "CrossEntropyLoss"  #:
SparseCrossEntropyLoss = "SparseCrossEntropyLoss"  #:
MSELoss = "MSELoss"  #:
# CTCLoss = "CTCLoss"  #:
L1Loss = "L1Loss"  #:
# NLLLoss = "NLLLoss"  #:
SmoothL1Loss = 'SmoothL1Loss'  #: Huber Loss or smooth L1 loss
SigmoidBCELoss = 'SigmoidBCELoss'  #: Binary Cross Entropy Loss
# KLDivLoss = 'KLDivLoss'  #: Killback-Leibler divergence loss
# CosineEmbeddingLoss = 'CosineEmbeddingLoss' #: Pytorch equivalent Cosine Embedding Loss
SoftmaxCrossEntropyLoss = 'SoftmaxCrossEntropyLoss'  #: Softmax + CE Loss together



import bson
import requests

from .common import getToken, read_in_chunks, MID_VERBOSITY, getVerbosity, get_progress_bar_uploader, convert_size
from .Task import Task
from .saving.saving import determine_model, hash_model
from .web.urls import RETRIEVE_MODEL_URL, HASH_URL, COMPILE_URL
from .saving import save_model


def compile(model, input_shape, library, model_label, asynchronous=False):
    if not isinstance(model_label, str):
        raise ValueError("Name given is not valid. Please supply a string.")
    if model_label != "" and model is None:
        print(model_label)
        params = {"token": getToken(), "input_shape": input_shape, "label": model_label}
        response = requests.get(RETRIEVE_MODEL_URL, params=params)
        if response.status_code == 200:
            return response
        else:
            raise LookupError("Model not found. " + str(response))
    if bson.ObjectId.is_valid(model) or isinstance(model, Task):
        return model
    if library == "":
        library = determine_model(model)
    hash = hash_model(model, library)
    params = {"token": getToken(), "input_shape": input_shape, "given_name": model_label, "hash": hash,
              "collection": 2, "modelType": library}
    response = requests.get(HASH_URL, params=params)
    if response.status_code == 200:
        if getVerbosity() >= MID_VERBOSITY:
            print("Model already on server. Returning result...")
        response = response.json()
        if isinstance(response, dict):
            message = response.get("message")
            print(message)
            response = response["id"]
        return response
    elif response.status_code != 204:
        raise ConnectionAbortedError("Checking hash not worked. {0}".format(response.content))
    else:
        if getVerbosity() >= MID_VERBOSITY:
            print("Model not found on server.")
            print("Saving model locally...")
        file_path = save_model(model, library)
        with open(file_path, "rb") as file:
            size = 0
            for piece in read_in_chunks(file):
                size += len(piece)
            if getVerbosity() >= MID_VERBOSITY:
                print("Model saved locally.")
            print("Model file size is {}. Uploading model now...".format(convert_size(size)))
            task = Task(compileApi(file, params))
            task.cache = {"model": model}
            if not asynchronous:
                task.wait()
                print("Model compiled successfully.")
            return task


def compileApi(file, params):
    file.seek(0)
    monitor = get_progress_bar_uploader(file)
    response = requests.post(COMPILE_URL, data=monitor, headers={'Content-Type': monitor.content_type}, params=params)
    if response.status_code != 200:
        raise ValueError(response.text)
    return response.json()




import os

BASE_URL = "https://api.neuro-ai.co.uk/"

if "BASE_URL" in os.environ:
    BASE_URL = os.environ["BASE_URL"]

COMPILE_URL = BASE_URL + "UploadModel"
RETRIEVE_MODEL_URL = BASE_URL + "RetrieveModel"
UPLOAD_DATA_URL = BASE_URL + "UploadData"
PREDICT_URL = BASE_URL + "Predict"
TRAIN_URL = BASE_URL + "Train"
HASH_URL = BASE_URL + "Hash"
TOKEN_URL = BASE_URL + "Token"
EXPORT_URL = BASE_URL + "Export"
TASK_STATUS_URL = BASE_URL + "status/"



"""
:synopsis: Global datasets available
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""

from ..core import Dataset

CIFAR10 = Dataset("cifar10", 40000, 50000, 60000)
""":CIFAR10: CIFAR10 dataset. Contains training/validation/test subsets of data. Accessed using 
`CIFAR10.train/val/test` properties. Can also get direct subset using slices. 
CIFAR10.train and CIFAR10.val contain the training set of the dataset with a split of 80% and 20% respectively.
CIFAR10.test contains 10000 samples from the non-training set of the dataset. 
"""

MNIST = Dataset("mnist", 48000, 60000, 70000)
""":MNIST: MNIST dataset. Contains training/validation/test subsets of data. Accessed using 
`MNIST.train/val/test` properties. Can also get direct subset using slices. 
MNIST.train and MNIST.val contain the training set of the dataset with a split of 80% and 20% respectively.
MNIST.test contains 10000 samples from the non-training set of the dataset. 
"""

KMNIST = Dataset("kmnist", 48000, 60000, 70000)
""":KMNIST: KMNIST dataset. Contains training/validation/test subsets of data. Accessed using 
`KMNIST.train/val/test` properties. Can also get direct subset using slices. 
KMNIST.train and KMNIST.val contain the training set of the dataset with a split of 80% and 20% respectively.
KMNIST.test contains 10000 samples from the non-training set of the dataset. 
"""

QMNIST = Dataset("qmnist", 48000, 60000, 120000)
""":QMNIST: QMNIST dataset. Contains training/validation/test subsets of data. Accessed using 
`QMNIST.train/val/test` properties. Can also get direct subset using slices. 
QMNIST.train and QMNIST.val contain the training set of the dataset with a split of 80% and 20% respectively.
QMNIST.test contains the 60000 samples from the non-training set of the dataset. 
"""

FashionMNIST = Dataset("fashion_mnist", 48000, 60000, 70000)
""":FashionMNIST: FashionMNIST dataset. Contains training/validation/test subsets of data. Accessed using 
`FashionMNIST.train/val/test` properties. Can also get direct subset using slices. 
FashionMNIST.train and FashionMNIST.val contain the training set of the dataset with a split of 80% and 20% respectively.
FashionMNIST.test contains the 10000 samples from the non-training set of the dataset. 
"""

CocoDetection = Dataset("coco_detection", 118287, 123287, 163957)
""":CocoDetection: CocoDetection dataset. Contains training/validation/test subsets of data. Accessed using 
`CocoDetection.train/val/test` properties. Can also get direct subset using slices. 
CocoDetection.train contains the 118287 samples from the train set of the dataset.
CocoDetection.val contains the 5000 samples from the val set of the dataset.
CocoDetection.test contains the 40670 samples from the test set of the dataset. 
"""




"""
:synopsis: Global models available. All pre-trained version of the following models have been trained using ImageNet.
    Input image expected (3, H, W) where H and W are at least 224. Images expected to have a range between 0 and 1 and
    normalized using mean = [0.485, 0.456, 0.406] and std = [0.229, 0.224, 0.225].
    Output vector of 1000 classes.
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""

from ..core import Model

alexnet = Model("alexnet")
""":alexnet: AlexNet model. Access pre-trained version `alexnet(pretrained=True)` """

vgg16 = Model("vgg16")
""":vgg16: VGG-16 model with batchnorm layers. Access pre-trained version `vgg16(pretrained=True)` """

vgg19 = Model("vgg19")
""":vgg19: VGG-19 model with batchnorm layers. Access pre-trained version `vgg19(pretrained=True)` """

resnet18 = Model("resnet18")
""":resnet18: ResNet-18 model. Access pre-trained version `resnet18(pretrained=True)` """

resnet34 = Model("resnet34")
""":resnet34: ResNet-34 model. Access pretrained version `resnet34(pretrained=True)` """

resnet50 = Model("resnet50")
""":resnet50: ResNet-50 model. Access pretrained version `resnet50(pretrained=True)` """

resnet101 = Model("resnet101")
""":resnet101: ResNet-101 model. Access pretrained version `resnet101(pretrained=True)` """

resnet152 = Model("resnet152")
""":resnet152: ResNet-152 model. Access pretrained version `resnet152(pretrained=True)` """

inception_v3 = Model("inception_v3")
""":inception_v3: Inception v3 model. Access pretrained version `inception_v3(pretrained=True)` """

resnext50_32x4d = Model("resnext50_32x4d")
""":resnext50_32x4d: ResNext-50 model. Access pretrained version `resnext50_32x4d(pretrained=True)` """

resnext101_32x8d = Model("resnext101_32x8d")
""":resnext101_32x8d: ResNext-101 model. Access pretrained version `resnext101_32x8d(pretrained=True)` """

mobilenet_v2 = Model("mobilenet_v2")
""":mobilenet_v2: MobileNet v2 model. Access pretrained version `mobilenet_v2(pretrained=True)` """

wide_resnet50_2 = Model("wide_resnet50_2")
""":wide_resnet50_2: Wide ResNet-50-2 model. Access pretrained version `wide_resnet50_2(pretrained=True)` """

wide_resnet101_2 = Model("wide_resnet101_2")
""":wide_resnet101_2: Wide ResNet-101-2 model. Access pretrained version `wide_resnet101_2(pretrained=True)` """

densenet161 = Model("densenet161")
""":densenet161: Densenet-161 model. Access pre-trained version `densenet161(pretrained=True)` """

googlenet = Model("googlenet")
""":googlenet: GoogleNet model. Access pre-trained version `googlenet(pretrained=True)` """

shufflenet_v2_x1_0 = Model("shufflenet_v2_x1_0")
""":shufflenet_v2_x1_0: ShuffleNetV2 model with 1.0x output channels. 
Access pre-trained version `shufflenet_v2_x1_0(pretrained=True)` """

mnasnet1_0 = Model("mnasnet1_0")
""":mnasnet1_0: MNASNet model. Access pre-trained version `mnasnet1_0(pretrained=True)` """






import base64

import dill
import requests

from .Dataset import Dataset
from .Task import Task
from .common import getToken, checkData, add_kwargs_to_params, checkModel, determine_data, validate_model, \
    check_model_type, check_data_type
from .web.urls import TRAIN_URL


def train(model, training_data, test_data, batch_size, epochs, optimiser, loss, trained_model_name, asynchronous, callback, **kwargs):
    checkModel(model)
    if isinstance(model, Task):
        cached_model = model.cache["model"]
        if not isinstance(training_data, (str, Dataset, dict)):
            validate_model(cached_model, training_data[0])
        if not isinstance(test_data, (str, Dataset, dict)) and test_data is not None:
            validate_model(cached_model, test_data[0])
    training_data = checkData(training_data)
    test_data = checkData(test_data)
    task_id = model.task_id if isinstance(model, Task) else ""
    task = Task(trainApi(model, training_data, test_data, optimiser, loss, batch_size, epochs, trained_model_name, task_id, **kwargs).json(),
                callback)
    if not asynchronous:
        task.wait(show_progress=True)
        print("Model finished training")
    return task


def trainApi(model, train_data, test_data, optimiser, loss, batch_size, epochs, trained_model_name, task_id="", **kwargs):
    if not isinstance(trained_model_name, str):
        raise ValueError("Name given is not valid. Please supply a string.")
    train_data, train_name, train_start, train_end = determine_data(train_data)
    test_data, test_name, test_start, test_end = determine_data(test_data)
    if callable(loss):
        print("Using custom loss function...")
        loss = base64.urlsafe_b64encode(dill.dumps(loss))
    params = { "loss": loss,
              "token": getToken(), "batch_size": batch_size, "epochs": epochs, "task_id": task_id, "train_start": train_start,
              "train_end": train_end, "test_start": test_start, "test_end": test_end, "train_name": train_name,
              "test_name": test_name, "trained_model_name": trained_model_name}
    check_model_type(model, params)
    check_data_type(train_data, "train", params)
    check_data_type(test_data, "test", params)
    params = add_kwargs_to_params(params, **kwargs)
    response = requests.get(TRAIN_URL, params=params, json=optimiser)
    if response.status_code != 200:
        raise ValueError(response.text)
    return response





"""
:synopsis: DataLoaders class wrapper
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""
import hashlib
import itertools

from npu.core.saving.saving import convert_to_numpy


class DataLoader(object):

    def __init__(self, dl):
        self.dl = dl
        self.shape = []

    def __str__(self):
        return self.dl

    def __iter__(self):
        return iter(self.dl)

    def __len__(self):
        return len(self.dl)

    def numpy(self):
        for data in self.dl:
            yield convert_to_numpy(data)

    def hash(self):
        data_part = next(self.numpy())
        if not self.shape:
            self.shape = [s.shape for s in data_part]
        else:
            for i, s in enumerate(self.shape):
                s[0] += data_part[i][0]

        def get_size(byte_list):
            return sum(len(d) for d in byte_list)

        def get_bytes(data):
            return [d.tobytes() for d in data]

        d_bytes = get_bytes(data_part)
        hashes = [hashlib.md5(d) for d in d_bytes]
        size = get_size(d_bytes)
        for data_part in itertools.islice(self.numpy(), 1, None):
            d_bytes = get_bytes(data_part)
            size += get_size(d_bytes)
            for i, d in enumerate(d_bytes):
                hashes[i].update(d)
        _hash = hashlib.md5()
        for __hash in hashes:
            _hash.update(__hash.digest())
        return _hash.hexdigest(), size







import requests

from .Dataset import Dataset
from .Task import Task
from .common import getToken, checkData, add_kwargs_to_params, checkModel, determine_data, validate_model, \
    check_model_type, check_data_type
from .web.urls import PREDICT_URL


def predict(model, data, asynchronous=False, callback=None, **kwargs):
    checkModel(model)
    if isinstance(model, Task) and model.cache is not None and "model" in model.cache:
        cached_model = model.cache["model"]
        if not isinstance(data, (str, Dataset, dict)):
            validate_model(cached_model, data)
    inference_data = checkData(data)
    task_id = model.task_id if isinstance(model, Task) else ""
    resp = predictApi(model, inference_data, task_id, **kwargs)
    task = Task(resp.json(), callback)
    if not asynchronous:
        task.wait()
    return task


def predictApi(model, data, task_id, **kwargs):
    data, data_name, start, end = determine_data(data)
    params = {"token": getToken(), "task_id": task_id, "start": start, "end": end,
              "data_name": data_name}
    check_model_type(model, params)
    check_data_type(data, "data", params)
    params = add_kwargs_to_params(params, **kwargs)
    response = requests.get(PREDICT_URL, params=params)
    if response.status_code != 200:
        raise ValueError(response.text)
    return response


